""" utils """
import boto3
from pkg_resources import resource_filename


def get_bucket_arn(bucket_name, region="", account_id=""):
    """ We need the arn of a queue """
    bucket_arn = f"arn:aws:s3:{region}:{account_id}:{bucket_name}"
    return bucket_arn


def get_aws_region_account():
    """ Return the aws region and account_id """
    session = boto3.Session()
    region = session.region_name
    sts = session.client("sts")
    response = sts.get_caller_identity()
    return region, response["Account"]


def get_resource(name):
    """ return resource content """
    path = resource_filename("duckdown.tool.provision", f"resources/{name}")
    with open(path) as file:
        return file.read()



""" save credentials """
import io
import os
import logging
import configparser
from contextlib import contextmanager
from .store import store

LOGGER = logging.getLogger(__name__)
DEFAULT_FILENAME = "credentials.cfg"


def _get_config_(default_path="~/.aws/credentials"):
    # open configparser
    config = configparser.ConfigParser(allow_no_value=True)
    config.read([f"./inventory/{DEFAULT_FILENAME}"])
    if not config:
        # read defaults and start over
        path = os.path.expanduser(default_path)
        assert os.path.isfile(path), "missing credentials"
        config.read([path])

    LOGGER.debug(config.sections())
    return config


def _save_config_(config):
    """ save to store """
    with io.StringIO() as stream:
        config.write(stream)
        credentials = stream.getvalue()

    cfg_path = store(DEFAULT_FILENAME, credentials, as_json=False)

    return cfg_path


def save_credentials(default_path="~/.aws/credentials"):
    """ open the default aws credentials and save in provision """
    config = _get_config_(default_path)
    LOGGER.debug(config.sections())
    return _save_config_(config)


def add_credentials(section, **kwargs):
    """ add kwargs to section """
    config = _get_config_()
    if not config.has_section(section):
        config.add_section(section)
    for option, value in kwargs.items():
        config.set(section, option, value)
    _save_config_(config)


@contextmanager
def using_credentials(section, default_path="~/.aws/credentials"):
    """ set the environ credentials """
    config = _get_config_(default_path)
    keys = ["AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY"]
    previous = {}
    for key in keys:
        previous[key] = os.environ.get(key, None)
        os.environ[key] = config[section][key]
    try:
        yield
    finally:
        for key, value in previous.items():
            if value:
                os.environ[key] = value
            else:
                del os.environ[key]



""" provision store """
import os
import logging
from duckdown.handlers.utils.json_utils import dumps

LOGGER = logging.getLogger(__name__)


def _make_inventory_store_():
    """ make a directory called inventory """
    # make inventory directory
    path = os.path.join(".", "inventory")
    if not os.path.exists(path):
        LOGGER.info("creating inventory directory: %s", path)
        os.makedirs(path, exist_ok=True)
    return path


def store(name, value, as_json=True):
    """ store file name with value in inventory """
    path = _make_inventory_store_()
    filename = f"{name}.json" if as_json else name
    file_path = os.path.join(path, filename)
    data = dumps(value, indent=4) if as_json else value
    with open(file_path, "w") as file:
        file.write(data)
    return file_path



""" iam tools """
import logging
import boto3
from .store import store
from .utils import get_resource, get_bucket_arn, get_aws_region_account
from . import credentials


LOGGER = logging.getLogger(__name__)


def policy_exists(policy_name):
    """ return PolicyArn if policy_name exists """
    # create client
    client = boto3.client("iam")
    response = client.list_policies(Scope="Local")

    LOGGER.debug(response)

    policy_names = {
        policy["PolicyName"]: policy["Arn"] for policy in response["Policies"]
    }

    return policy_names.get(policy_name)


def get_policy(policy_arn):
    """ get policy body of policy_arn """
    # create client and get policy
    client = boto3.client("iam")
    response = client.get_policy(PolicyArn=policy_arn)

    LOGGER.debug(response)

    policy = response["Policy"]
    policy_version = policy["DefaultVersionId"]
    # get default policy version
    response = client.get_policy_version(
        PolicyArn=policy_arn, VersionId=policy_version
    )

    LOGGER.debug(response)
    # store result
    policy_doc = response["PolicyVersion"]["Document"]
    policy_name = policy_arn.split("/")[-1]
    store(f"policy-{policy_name}.json", policy_doc)

    return policy_doc


def _get_access_keys_(client, user_name):
    """ fetch a user access_keys """
    # create client and get policy
    response = client.list_access_keys(UserName=user_name)

    LOGGER.debug(response)
    return [
        item
        for item in response["AccessKeyMetadata"]
        if item["Status"] == "Active"
    ]


def _get_attached_policies_(client, user_name):
    # fetch attached_policies
    response = client.list_attached_user_policies(UserName=user_name)
    return response["AttachedPolicies"]


def get_user(user_name):
    """ fetch a user and their policies """
    # create client and get policy
    client = boto3.client("iam")
    response = client.get_user(UserName=user_name)

    LOGGER.debug(response)

    user = response["User"]

    # fetch attached_policies
    response = client.list_attached_user_policies(UserName=user_name)
    user["policies"] = _get_attached_policies_(client, user_name)
    user["access_keys"] = _get_access_keys_(client, user_name)

    store(f"user-{user_name}.json", user)

    return user


def create_user(user_name, bucket_name):
    """ create an iam user """
    region, _ = get_aws_region_account()
    bucket_arn = get_bucket_arn(bucket_name)

    # create client and get policy
    client = boto3.client("iam")
    response = client.create_user(
        UserName=user_name, Tags=[{"Key": "app", "Value": "duckdown"}]
    )
    LOGGER.debug(response)

    # set policy
    user_policy = get_resource("user_policy_tmpl.json").format(
        bucket_arn=bucket_arn
    )
    policy_name = f"{user_name}-duckdown"
    response = client.create_policy(
        PolicyName=policy_name,
        PolicyDocument=user_policy,
        Description="upload to duckdown",
    )
    LOGGER.debug(response)

    policy_arn = response["Policy"]["Arn"]
    response = client.attach_user_policy(
        UserName=user_name, PolicyArn=policy_arn
    )

    # set access keys
    response = client.create_access_key(UserName=user_name)
    LOGGER.debug(response)
    credentials.add_credentials(
        user_name,
        aws_access_key_id=response["AccessKey"]["AccessKeyId"],
        aws_secret_access_key=response["AccessKey"]["SecretAccessKey"],
        region=region,
    )

    return response


def delete_user(user_name):
    """ delete an iam user """
    # get the user
    user = get_user(user_name)

    # create a client
    client = boto3.client("iam")

    # Attached managed policies ( DetachUserPolicy )
    for policy in user["policies"]:
        policy_arn = policy["PolicyArn"]
        response = client.detach_user_policy(
            UserName=user_name, PolicyArn=policy_arn
        )
        LOGGER.debug(response)
        response = client.delete_policy(PolicyArn=policy_arn)
        LOGGER.debug(response)

    # Access keys ( DeleteAccessKey )
    for key in user["access_keys"]:
        key_id = key["AccessKeyId"]
        response = client.delete_access_key(
            UserName=user_name, AccessKeyId=key_id
        )
        LOGGER.debug(response)

    # delete the user
    response = client.delete_user(UserName=user_name)
    LOGGER.debug(response)
    return response



""" provision tasks """
from invoke import task
from duckdown.handlers.utils import json_utils
from .credentials import using_credentials
from . import s3_tools, iam_tools, route53_tools


@task
def get_region(_):
    """ aws region """
    print(s3_tools.get_aws_region_account())


@task
def get_buckets(_):
    """ list buckets """
    response = s3_tools.buckets()
    print(json_utils.dumps(response, indent=4))


@task
def get_bucket(_, name):
    """ get bucket """
    response = s3_tools.get_bucket(name)
    print(json_utils.dumps(response, indent=4))


@task
def create_bucket(_, name):
    """ create a bucket on s3 """
    response = s3_tools.create_bucket(name, True)
    print(json_utils.dumps(response, indent=4))


@task
def delete_bucket(_, name):
    """ delete bucket on s3 """
    response = s3_tools.delete_bucket(name)
    print(json_utils.dumps(response, indent=4))


@task
def create_website(_, name, title=None):
    """ convert bucket to s3 website """
    response = s3_tools.make_website(name, title)
    print(json_utils.dumps(response, indent=4))


@task
def upload(_, creds, bucket_name, key, filepath):
    """ upload file to bucket """
    with using_credentials(creds):
        response = s3_tools.upload(bucket_name, key, filepath)
        print(json_utils.dumps(response, indent=4))


@task
def get_user(_, name):
    """ get iam user """
    response = iam_tools.get_user(name)
    print(json_utils.dumps(response, indent=4))


@task
def create_user(_, name, bucket_name):
    """ create a user on iam """
    response = iam_tools.create_user(name, bucket_name)
    print(json_utils.dumps(response, indent=4))


@task
def delete_user(_, name):
    """ delete user on iam """
    response = iam_tools.delete_user(name)
    print(json_utils.dumps(response, indent=4))


@task
def get_domains(_):
    """ list current domains """
    response = route53_tools.get_domains()
    print(json_utils.dumps(response, indent=4))


@task
def get_hosted_zones(_):
    """ list current domains """
    response = route53_tools.get_hosted_zones()
    print(json_utils.dumps(response, indent=4))


@task
def set_bucket_cname(_, bucket_name):
    """ set the CNAME on the domain hosted zone in route53 """
    response = route53_tools.set_website_cname(bucket_name)
    print(json_utils.dumps(response, indent=4))



""" Mixin to perform asset_url """
from tornado.web import StaticFileHandler


class AssetsMixin:  # pylint: disable=too-few-public-methods
    """ mixin to add assert_url to handler """

    def asset_url(self, path):
        """ return asset handler versioned url """
        return StaticFileHandler.make_static_url(
            {
                "static_url_prefix": "/edit/assets/",
                "static_path": self.application.settings["duck_assets"],
            },
            path,
        )



# pylint: disable=W0201
""" Handle request for index page """
from tornado.web import RequestHandler, authenticated
from .access_control import UserMixin
from .utils.assets_mixin import AssetsMixin


class EditorHandler(
    AssetsMixin, UserMixin, RequestHandler
):  # pylint: disable=W0223
    """ return index page """

    def initialize(self, manifest, page):
        """ setup production """
        self.manifest = manifest
        self.page = page

    @property
    def duck_path(self):
        """ return application duck_path """
        return self.application.settings.get("duck_path")

    @property
    def img_path(self):
        """ return application img_path """
        if self.application.settings.get("local_images"):
            path = self.application.settings.get("img_path", "")
            return f"/static/{path}"
        return self.application.settings.get("img_path")

    def get_template_path(self):
        """ return app resource """
        return self.application.settings["duck_templates"]

    @authenticated
    def get(self):
        """ handle get request """
        self.render(self.page)



""" route53 tools """
import logging
import boto3
from tld import get_tld
from .store import store
from .utils import get_aws_region_account

LOGGER = logging.getLogger(__name__)


def check_available(domain):
    """ check the availability of domain with route53 """

    # check with route53
    client = boto3.client("route53domains", region_name="us-east-1")
    availability = client.check_domain_availability(DomainName=domain)

    LOGGER.debug(availability)

    # store in provision
    availability["domain"] = domain
    store(domain, availability)

    return availability


def get_domains():
    """ list domains """
    client = boto3.client("route53", region_name="us-east-1")
    response = client.get_domains()
    LOGGER.debug(response)
    domains = response["Domains"]
    store("domains", domains)
    return response


def get_hosted_zones():
    """ list domains """
    client = boto3.client("route53", region_name="us-east-1")
    response = client.list_hosted_zones()
    LOGGER.debug(response)
    zones = response["HostedZones"]
    store("zones", zones)
    return zones


def get_hosted_zone_id(name):
    """ return the id of hosted zone """
    # for some reason domain hosted zone end with .
    key = f"{name}."
    for zone in get_hosted_zones():
        print(zone)
        if zone["Name"] == key:
            return zone["Id"]
    raise ValueError(f"No zone called {name}")


def _extract_domain_(bucket_name):
    """ return the domain part of bucket_name """
    res = get_tld(f"http://{bucket_name}", as_object=True)
    return res.fld


def set_website_cname(bucket_name):
    """ it is assumed that you bucket is a subdomain """

    region, _ = get_aws_region_account()

    domain = _extract_domain_(bucket_name)

    hosted_zone_id = get_hosted_zone_id(domain)

    target = f"http://{bucket_name}.s3-website-{region}.amazonaws.com"

    client = boto3.client("route53", region_name="us-east-1")
    change_batch = {
        "Comment": "add %s -> %s" % (bucket_name, target),
        "Changes": [
            {
                "Action": "UPSERT",
                "ResourceRecordSet": {
                    "Name": bucket_name,
                    "Type": "CNAME",
                    "TTL": 300,
                    "ResourceRecords": [{"Value": target}],
                },
            }
        ],
    }
    store(f"cname-{bucket_name}", change_batch)
    response = client.change_resource_record_sets(
        HostedZoneId=hosted_zone_id, ChangeBatch=change_batch
    )
    LOGGER.debug(response)
    return response


def request_acm_certificate(bucket_name, transaction):
    """ request an ssl certificate from ACM """

    client = boto3.client("ACM")
    response = client.request_certificate(
        DomainName=bucket_name,
        ValidationMethod="DNS",
        IdempotencyToken=transaction,
        DomainValidationOptions=[
            {"DomainName": "string", "ValidationDomain": "string"},
        ],
        Options={"CertificateTransparencyLoggingPreference": "ENABLED"},
        CertificateAuthorityArn="string",
    )
    LOGGER.debug(response)
    return response



""" tools for us with s3 """
import logging
import mimetypes
import boto3
from botocore.exceptions import ClientError
from pkg_resources import resource_filename
from tornado.template import Loader
from duckdown.handlers.utils.json_utils import loads
from .store import store
from .utils import get_resource, get_aws_region_account, get_bucket_arn

LOGGER = logging.getLogger(__name__)


def buckets():
    """ return a set of bucket names """
    # create s3 client
    client = boto3.client("s3")
    response = client.list_buckets()
    LOGGER.debug(response)
    return response["Buckets"]


def bucket_exists(bucket_name):
    """ check to see if bucket exists """
    return bucket_name in {bucket["Name"] for bucket in buckets()}


def get_bucket(bucket_name):
    """ get a bucket and store provision """
    result = None
    for bucket in buckets():
        if bucket["Name"] == bucket_name:
            result = bucket
            break

    if result:
        # create s3 client
        client = boto3.client("s3")

        result["arn"] = get_bucket_arn(bucket_name)

        # load cors, acl and policy
        response = client.get_bucket_cors(Bucket=bucket_name)
        LOGGER.debug(response)
        result["cors"] = response["CORSRules"]

        response = client.get_bucket_acl(Bucket=bucket_name)
        LOGGER.debug(response)
        result["acl"] = {
            "Owner": response["Owner"],
            "Grants": response["Grants"],
        }

        response = client.get_bucket_policy(Bucket=bucket_name)
        LOGGER.debug(response)
        result["policy"] = loads(response["Policy"])

        response = client.get_bucket_location(Bucket=bucket_name)
        LOGGER.debug(response)
        result["location"] = response["LocationConstraint"]

        try:
            response = client.get_bucket_website(Bucket=bucket_name)
            LOGGER.debug(response)
            result["website"] = {
                key: value
                for key, value in response.items()
                if key != "ResponseMetadata"
            }
        except ClientError as ex:
            if "NoSuchWebsiteConfiguration" not in str(ex):
                raise

        # store
        store(f"bucket-{bucket_name}", result)
    return result


def create_bucket(bucket_name, public=False, region=None):
    """ create a bucket """
    if region is None:
        region, _ = get_aws_region_account()

    # create s3 client
    client = boto3.client("s3")

    # create bucket
    LOGGER.info("creating bucket: %r in %r", bucket_name, region)

    args = {
        "ACL": "public-read" if public else "private",
        "Bucket": bucket_name,
        "CreateBucketConfiguration": {"LocationConstraint": region},
    }
    LOGGER.info(args)
    response = client.create_bucket(**args)
    LOGGER.debug(response)

    arn = get_bucket_arn(bucket_name)

    # set cors
    cors = {"CORSRules": loads(get_resource("cors_tmpl.json"))}
    response = client.put_bucket_cors(
        Bucket=bucket_name, CORSConfiguration=cors
    )
    LOGGER.debug(response)

    # set policy
    bucket_policy = get_resource("bucket_policy_tmpl.json").format(
        bucket_arn=arn
    )

    response = client.put_bucket_policy(
        Bucket=bucket_name, Policy=bucket_policy
    )
    LOGGER.debug(response)

    return get_bucket(bucket_name)


def delete_bucket(bucket_name):
    """ delete a bucket """
    # create s3 client
    client = boto3.client("s3")

    # get contents
    response = client.list_objects(Bucket=bucket_name)
    LOGGER.debug(response)

    # remove contents
    for doc in response.get("Contents", []):
        client.delete_object(Bucket=bucket_name, Key=doc["Key"])

    # delete bucket
    response = client.delete_bucket(Bucket=bucket_name)
    LOGGER.debug(response)

    return response


def upload(bucket_name, key, file_path):
    """ upload object to bucket """
    # create s3 client
    client = boto3.client("s3")
    mime, _ = mimetypes.guess_type(file_path)
    with open(file_path, "rb") as file:
        response = client.put_object(
            Body=file, Bucket=bucket_name, Key=key, ContentType=mime
        )
    LOGGER.debug(response)

    return response


def make_website(bucket_name, title=None):
    """ make a website and upload index and error pages """
    # region for url
    region, _ = get_aws_region_account()

    # create s3 client
    client = boto3.client("s3")

    if title is None:
        title = bucket_name

    # load pages
    loader = Loader(resource_filename("duckdown.tool.provision", "resources"))
    index_page = loader.load("index_tmpl.html").generate(title=title)
    error_page = loader.load("error_tmpl.html").generate(title=title)
    response = client.put_object(
        Body=index_page,
        Bucket=bucket_name,
        Key="index.html",
        ContentType="text/html",
    )
    LOGGER.debug(response)
    response = client.put_object(
        Body=error_page,
        Bucket=bucket_name,
        Key="error.html",
        ContentType="text/html",
    )
    LOGGER.debug(response)

    # set as website
    response = client.put_bucket_website(
        Bucket=bucket_name,
        WebsiteConfiguration={
            "ErrorDocument": {"Key": "error.html"},
            "IndexDocument": {"Suffix": "index.html"},
        },
    )
    LOGGER.debug(response)

    get_bucket(bucket_name)
    return f"http://{bucket_name}.s3-website-{region}.amazonaws.com"



# pylint: disable=unused-variable, too-many-locals
""" application entry point """
import json
import logging
import os
import tornado.ioloop
import tornado.web
import tornado.log
import convoke
from dotenv import load_dotenv
from pkg_resources import resource_filename

from .handlers import EditorHandler
from .handlers import SiteHandler
from .handlers import DirHandler
from .handlers import MarkHandler
from .handlers import S3Browser
from .handlers import LoginHandler, LogoutHandler
from .handlers.utils.s3tmpl_loader import S3Loader

LOGGER = logging.getLogger(__name__)


def make_app_path(settings, name, default):
    """ return path relative to app """
    return os.path.join(
        settings.get("app_path", ""), settings.get(name, default)
    )


def make_app():
    """ make a tornado application """
    settings = convoke.get_settings("duckdown")
    debug = settings.as_bool("debug", default="False")
    production = settings.as_bool("production", default="True")

    if production is False:
        manifest = None
    else:
        LOGGER.info("loading client manifest")
        with open(
            resource_filename("duckdown", "assets/vue/manifest.json")
        ) as file:
            manifest = json.load(file)

    static_path = make_app_path(settings, "static_path", "static")
    template_path = make_app_path(settings, "template_path", "templates")
    pages_path = make_app_path(settings, "pages_path", "pages")
    users_path = make_app_path(settings, "users_path", "users.json")
    img_path = settings.get("img_path", "")

    tornado_settings = {
        "debug": debug,
        "production": production,
        "port": settings.as_int("port", default="8080"),
        "duck_users": users_path,
        "duck_path": "/edit/assets/",
        "duck_assets": resource_filename("duckdown", "assets"),
        "duck_templates": resource_filename("duckdown", "templates"),
        "static_path": settings.get("static", static_path),
        "template_path": settings.get("templates", template_path),
        "cookie_name": settings.get("cookie_name", "duckdown-cookie"),
        "cookie_secret": settings.get(
            "cookie_secret", "it was a dark and stormy duckdown"
        ),
        "login_url": settings.get("login_url", "/login"),
        "app_name": settings.get("app_name", "duckdown-app"),
        "local_images": settings.as_bool("local_images", default="False"),
        "img_path": img_path,
        "compress_response": True,
    }

    # load aws credentials
    image_bucket = {
        "bucket_name": settings.get("image_bucket", None),
        "aws_access_key_id": settings.get("AWS_ACCESS_KEY_ID", None),
        "aws_secret_access_key": settings.get("AWS_SECRET_ACCESS_KEY", None),
        "folder": "images/",
    }

    LOGGER.info("settings:")
    for key, value in tornado_settings.items():
        LOGGER.info("\t%s: %r", key, value)

    s3_loader = S3Loader(
        **{
            "bucket_name": settings.get("image_bucket", ""),
            "aws_access_key_id": settings.get("AWS_ACCESS_KEY_ID", ""),
            "aws_secret_access_key": settings.get("AWS_SECRET_ACCESS_KEY", ""),
            "folder": "templates",
        }
    )

    routes = [
        (r"/login", LoginHandler, {"users": json.load(open(users_path))}),
        (r"/logout", LogoutHandler),
        (r"/browse/(.*)", S3Browser, image_bucket),
        (
            r"/edit/assets/(.*)",
            tornado.web.StaticFileHandler,
            {"path": tornado_settings["duck_assets"]},
        ),
        (r"/edit/mark/", MarkHandler),
        (r"/edit/pages/(.*)", DirHandler, {"directory": pages_path}),
        (r"/edit", EditorHandler, {"manifest": manifest, "page": "vue.html"}),
        (r"/(.*)", SiteHandler, {"pages": pages_path, "s3_loader": None}),
    ]

    if production is False:
        LOGGER.info("installing vue dev handler")
        routes.insert(
            0,
            (
                r"/src/(.*)",
                tornado.web.StaticFileHandler,
                {"path": "./client/src/"},
            ),
        )
    else:
        LOGGER.info("installing vue handler")
        _assets = resource_filename("duckdown", "assets/vue/")
        routes.insert(
            0,
            (
                r"/_assets/(.*)",
                tornado.web.StaticFileHandler,
                {"path": _assets},
            ),
        )

    return tornado.web.Application(routes, **tornado_settings)


def main():
    """ make an app and run it """
    load_dotenv(verbose=True)
    app = make_app()

    app.listen(app.settings["port"])
    LOGGER.info("listening on port: %s", app.settings["port"])

    if app.settings["debug"] is True:
        LOGGER.info("running in debug mode")

    ioloop = tornado.ioloop.IOLoop.current()
    try:
        ioloop.start()
    except KeyboardInterrupt:
        LOGGER.info("shutting down")
        ioloop.stop()


if __name__ == "__main__":
    tornado.log.enable_pretty_logging()
    main()



""" A utility to scan a directory and generate navigation """
import os
import logging

LOGGER = logging.getLogger(__name__)


def parse_header(content, tag=None, alt_tag=None):
    """ return header or header tag value """
    values = {}
    parts = content.split("\n\n", 1)
    if len(parts) == 2:
        header, _ = parts
        for line in header.split("\n"):
            try:
                name, value = line.split(":", maxsplit=1)
            except ValueError:
                LOGGER.info(line)
                raise
            values[name.strip()] = value.strip()
    result = None
    if tag:
        result = values.get(tag)
    if result is None and alt_tag is not None:
        result = values.get(alt_tag)
    if result is None:
        result = values
    return result


def nav(root: str, path: str) -> str:  # pylint: disable=W0613
    """
    walk from root and discover index.md
    grab the title and generate <ul class="nav>
    go as deep a peers to path
    """
    result = []
    result.append('<ul class="nav">')
    indent = "    "
    depth = 0
    for dirpath, _, filenames in os.walk(root):
        for filename in filenames:
            if filename == "index.md":
                file_path = os.path.join(dirpath, filename)
                with open(file_path) as file:
                    title = parse_header(file.read(), "nav", "Title")
                if title:
                    fpath, _ = os.path.splitext(file_path)
                    rel_path = os.path.relpath(f"{fpath}.html", root)
                    new_depth = rel_path.count("/")
                    if new_depth > depth:
                        result.append(f'{indent * (depth+1)}<ul class="nav">')
                    elif new_depth < depth:
                        result.append(f"{indent * depth}</ul>")
                    result.append(
                        f'{indent * (new_depth+1)}<li><a href="/{rel_path}">{title}</a></li>'
                    )
                    depth = new_depth
    while depth:
        result.append(f"{indent * depth}</ul>")
        depth = depth - 1
    result.append("</ul>")
    return result
